@@ -4,7 +4,7 @@ import pathlib
 import platform
 import sys
 from dataclasses import dataclass
-from typing import Callable, List
+from typing import Any, Callable, List
 
 if sys.version_info >= (3, 11):
     from typing import Self
@@ -15,8 +15,8 @@ import pyfwfinder as fwf
 from result import Err, Ok, Result
 
 from freewili.framing import ResponseFrame
-from freewili.serial_util import FreeWiliSerial
-from freewili.types import ButtonColor, FreeWiliProcessorType, IOMenuCommand
+from freewili.fw_serial import FreeWiliSerial
+from freewili.types import ButtonColor, EventType, FreeWiliProcessorType, IOMenuCommand
 
 # USB Locations:
 # first address = FTDI
@@ -26,24 +26,6 @@ DISPLAY_HUB_LOC_INDEX = 1
 # third address = Main
 MAIN_HUB_LOC_INDEX = 0
 
-# This maps the actual GPIO exposed on the connector, all others not
-# listed here are internal to the processor.
-GPIO_MAP = {
-    8: "GPIO8/UART1_Tx_OUT",
-    9: "GPIO9/UART1_Rx_IN",
-    10: "GPIO10/UART1_CTS_IN",
-    11: "GPIO11/UART1_RTS_OUT",
-    12: "GPIO12/SPI1_Rx_IN",
-    13: "GPIO13/SPI1_CS_OUT",
-    14: "GPIO14/SPI1_SCLK_OUT",
-    15: "GPIO15/SPI1_Tx_OUT",
-    16: "GPIO16/I2C0 SDA",
-    17: "GPIO17/I2C0 SCL",
-    25: "GPIO25/GPIO25_OUT",
-    26: "GPIO26/GPIO26_IN",
-    27: "GPIO27/GPIO_27_OUT",
-}
-
 
 class FreeWili:
     """Free-Wili device used to access FTDI and serial functionality."""
@@ -61,6 +43,29 @@ class FreeWili:
     def __repr__(self) -> str:
         return f"<{self.__class__.__name__}: {self.device.serial}>"
 
+    def __enter__(self) -> Self:
+        self.open().expect("Failed to open FreeWili")
+        return self
+
+    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
+        self.close()
+
+    def set_event_callback(self, event_cb: None | Callable[[EventType, ResponseFrame, Any], None]) -> None:
+        """Set the event callback for the FreeWili.
+
+        Parameters:
+        -----------
+            event_cb: Callable[[EventType, ResponseFrame, Any], None]
+                Callback function to be called when an event occurs.
+                The function should accept three arguments: the event type, the response frame,
+                and any additional data.
+                Set to None to disable the callback.
+        """
+        if self.main_serial:
+            self.main_serial.set_event_callback(event_cb)
+        if self.display_serial:
+            self.display_serial.set_event_callback(event_cb)
+
     @property
     def usb_devices(self) -> List[fwf.USBDevice]:
         """Grab all the USB devices attached to the FreeWili."""
@@ -70,12 +75,12 @@ class FreeWili:
         """Get the USBDevice by ProcessorType.
 
         Parameters:
-        ----------
+        ------------
             processor_type: FreeWiliProcessorType
                 Type of processor to get.
 
         Returns:
-        -------
+        ---------
             None | fwf.USBDevice:
                 fwf.USBDevice if successful, None otherwise.
 
@@ -157,7 +162,7 @@ class FreeWili:
             None
 
         Returns:
-        --------
+        ---------
             None | FreeWiliSerial:
                 FreeWiliSerial on success, None otherwise.
         """
@@ -176,7 +181,7 @@ class FreeWili:
             None
 
         Returns:
-        --------
+        ---------
             None | FreeWiliSerial:
                 FreeWiliSerial on success, None otherwise.
         """
@@ -195,7 +200,7 @@ class FreeWili:
                 Processor type to get serial port for.
 
         Returns:
-        --------
+        ---------
             Result[FreeWiliSerial, str]:
                 Ok(FreeWiliSerial) on success, Err(str) otherwise.
         """
@@ -237,7 +242,7 @@ class FreeWili:
                 number of seconds to wait when blocking.
 
         Returns:
-        -------
+        ---------
             Result[None, str]:
                 Ok(None) if successful, Err(str) otherwise.
         """
@@ -260,7 +265,7 @@ class FreeWili:
                 Restore the menu functionality before close
 
         Returns:
-        -------
+        ---------
             None
         """
         if self.main_serial:
@@ -273,11 +278,11 @@ class FreeWili:
         """Find first Free-Wili device attached to the host.
 
         Parameters:
-        ----------
+        ------------
             None
 
         Returns:
-        -------
+        ---------
             Result[FreeWili, str]:
                 Ok(FreeWili) if successful, Err(str) otherwise.
 
@@ -298,11 +303,11 @@ class FreeWili:
         """Find all Free-Wili devices attached to the host.
 
         Parameters:
-        ----------
+        ------------
             None
 
         Returns:
-        -------
+        ---------
             tuple[FreeWili, ...]:
                 Tuple of FreeWili devices.
 
@@ -341,7 +346,7 @@ class FreeWili:
                 Size of the chunks to send in bytes. Typically this should be left at the default value.
 
         Returns:
-        -------
+        ---------
             Result[str, str]:
                 Returns Ok(str) if the command was sent successfully, Err(str) if not.
         """
@@ -386,7 +391,7 @@ class FreeWili:
                     def user_callback(msg: str) -> None
 
         Returns:
-        -------
+        ---------
             Result[str, str]:
                 Returns Ok(str) if the command was sent successfully, Err(str) if not.
         """
@@ -417,7 +422,7 @@ class FreeWili:
                 Processor to upload the file to.
 
         Returns:
-        -------
+        ---------
             Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
         """
@@ -433,12 +438,12 @@ class FreeWili:
         """Get all the IO values.
 
         Parameters:
-        ----------
+        ------------
             processor: FreeWiliProcessorType
                 Processor to set IO on.
 
         Returns:
-        -------
+        ---------
             Result[tuple[int], str]:
                 Ok(tuple[int]) if the command was sent successfully, Err(str) if not.
         """
@@ -450,6 +455,31 @@ class FreeWili:
             case _:
                 raise RuntimeError("Missing case statement")
 
+    def toggle_high_speed_io(
+        self: Self,
+        enable: bool,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
+    ) -> Result[str, str]:
+        """Toggle the high-speed Bidirectional IO.
+
+        Parameters:
+        ------------
+            enable: bool
+                Whether to enable or disable high-speed IO.
+
+        Returns:
+        ---------
+            Result[ResponseFrame, str]:
+                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.toggle_high_speed_io(enable)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
     def set_io(
         self: Self,
         io: int,
@@ -461,7 +491,7 @@ class FreeWili:
         """Set the state of an IO pin to high or low.
 
         Parameters:
-        ----------
+        ------------
             io : int
                 The number of the IO pin to set.
             menu_cmd : IOMenuCommand
@@ -474,7 +504,7 @@ class FreeWili:
                 Processor to set IO on.
 
         Returns:
-        -------
+        ---------
             Result[ResponseFrame, str]:
                 Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
         """
@@ -497,7 +527,7 @@ class FreeWili:
         """Set the GUI RGB LEDs.
 
         Parameters:
-        ----------
+        ------------
             io : int
                 The number of the IO pin to set.
             red : int
@@ -510,7 +540,7 @@ class FreeWili:
                 Processor to set LEDs on.
 
         Returns:
-        -------
+        ---------
             Result[ResponseFrame, str]:
                 Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
         """
@@ -528,7 +558,7 @@ class FreeWili:
         """Write I2C data.
 
         Parameters:
-        ----------
+        ------------
             address : int
                 The address to write to.
             register : int
@@ -539,7 +569,7 @@ class FreeWili:
                 Processor to use.
 
         Returns:
-        -------
+        ---------
             Result[ResponseFrame, str]:
                 Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
         """
@@ -557,7 +587,7 @@ class FreeWili:
         """Write I2C data.
 
         Parameters:
-        ----------
+        -----------
             address : int
                 The address to write to.
             register : int
@@ -568,7 +598,7 @@ class FreeWili:
                 Processor to use.
 
         Returns:
-        -------
+        ---------
             Result[str, str]:
                 Ok(str) if the command was sent successfully, Err(str) if not.
                 The str is the response from the device, typically "OK".
@@ -585,12 +615,12 @@ class FreeWili:
         """Poll I2C data.
 
         Parameters:
-        ----------
+        ------------
             processor: FreeWiliProcessorType
                 Processor to use.
 
         Returns:
-        -------
+        ---------
             Result[tuple[int, ...], str]:
                 Ok(tuple[int, ...]) if the command was sent successfully, Err(str) if not.
         """
@@ -615,7 +645,7 @@ class FreeWili:
                 Processor to use.
 
         Returns:
-        -------
+        ---------
             Result[ResponseFrame, str]:
                 Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
         """
@@ -640,7 +670,7 @@ class FreeWili:
                 Processor to use.
 
         Returns:
-        -------
+        ---------
             Result[ResponseFrame, str]:
                 Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
         """
@@ -663,7 +693,7 @@ class FreeWili:
                 Processor to use.
 
         Returns:
-        -------
+        ---------
             Result[dict[ButtonColor, bool], str]:
                 Ok(dict[ButtonColor, bool]) if the command was sent successfully, Err(str) if not.
         """
@@ -686,7 +716,7 @@ class FreeWili:
                 Processor to use.
 
         Returns:
-        -------
+        ---------
             Result[ResponseFrame, str]:
                 Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
         """
@@ -698,6 +728,365 @@ class FreeWili:
             case _:
                 raise RuntimeError("Missing case statement")
 
+    def enable_accel_events(
+        self,
+        enable: bool,
+        interval_ms: int | None = None,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
+    ) -> Result[str, str]:
+        """Enable or disable acceleration events.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable acceleration events.
+            interval_ms: int | None
+                The interval in milliseconds for accelerometer events. If None, the default value will be used.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[ResponseFrame, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.enable_accel_events(enable, interval_ms)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def enable_gpio_events(
+        self,
+        enable: bool,
+        interval_ms: int | None = None,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
+    ) -> Result[str, str]:
+        """Enable or disable GPIO events.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable GPIO events.
+            interval_ms: int | None
+                The interval in milliseconds for GPIO events. If None, the default value will be used.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[ResponseFrame, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.enable_gpio_events(enable, interval_ms)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def enable_button_events(
+        self,
+        enable: bool,
+        interval_ms: int | None = None,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
+    ) -> Result[str, str]:
+        """Enable or disable acceleration events.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable acceleration events.
+            interval_ms: int | None
+                The interval in milliseconds for accelerometer events. If None, the default value will be used.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[ResponseFrame, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.enable_button_events(enable, interval_ms)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def enable_ir_events(
+        self,
+        enable: bool,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
+    ) -> Result[str, str]:
+        """Enable or disable infrared events.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable infrared events.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[ResponseFrame, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.enable_ir_events(enable)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def enable_battery_events(
+        self,
+        enable: bool,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
+    ) -> Result[str, str]:
+        """Enable or disable battery events.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable battery events.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[ResponseFrame, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.enable_battery_events(enable)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def enable_radio_events(
+        self,
+        enable: bool,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
+    ) -> Result[str, str]:
+        """Enable or disable radio events on currently selected radio.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable radio events.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[ResponseFrame, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.enable_radio_events(enable)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def enable_uart_events(
+        self,
+        enable: bool,
+        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
+    ) -> Result[str, str]:
+        """Enable or disable UART events.
+
+        Arguments:
+        ----------
+            enable: bool
+                Whether to enable or disable UART events.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[ResponseFrame, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.enable_uart_events(enable)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def process_events(self) -> None:
+        """Process any events that have been received.
+
+        This is typically called in a loop to process events as they come in.
+        """
+        if self.main_serial:
+            self.main_serial.process_events()
+        if self.display_serial:
+            self.display_serial.process_events()
+
+    def select_radio(
+        self, radio_index: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[str, str]:
+        """Select the radio to use for events.
+
+        Arguments:
+        ----------
+            radio_index: int
+                Index of the radio to select. 1 or 2 typically.
+
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.select_radio(radio_index)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def set_radio_event_rssi_threshold(
+        self, rssi: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[str, str]:
+        """Set the RSSI threshold for radio events.
+
+        Arguments:
+        ----------
+            rssi: int
+                RSSI threshold value to set.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.set_radio_event_rssi_threshold(rssi)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def set_radio_event_sample_window(
+        self, sample_window_ms: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[str, str]:
+        """Set the sample window (ms) for the specified radio.
+
+        Arguments:
+        ----------
+            sample_window_ms: int
+                Sample window value to set.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        ---------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.set_radio_event_sample_window(sample_window_ms)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def transmit_radio_subfile(
+        self, sub_fname: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[str, str]:
+        """Transmit a subfile to the specified radio.
+
+        Arguments:
+        ----------
+            sub_fname: str
+                Name of the subfile to transmit. This should be the filename with the extension.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.transmit_radio_subfile(sub_fname)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def write_radio(
+        self, data: bytes, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[str, str]:
+        """Write radio data.
+
+        Arguments:
+        ----------
+            data: bytes
+                Data to transmit.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.write_radio(data)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
+    def write_uart(
+        self, data: bytes, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
+    ) -> Result[str, str]:
+        """Write uart data.
+
+        Arguments:
+        ----------
+            data: bytes
+                Data to transmit.
+            processor: FreeWiliProcessorType
+                Processor to use.
+
+        Returns:
+        -------
+            Result[str, str]:
+                Ok(str) if the command was sent successfully, Err(str) if not.
+        """
+        match self.get_serial_from(processor):
+            case Ok(serial):
+                return serial.write_uart(data)
+            case Err(msg):
+                return Err(msg)
+            case _:
+                raise RuntimeError("Missing case statement")
+
     def reset_to_uf2_bootloader(self, processor: FreeWiliProcessorType) -> Result[None, str]:
         """Reset the FreeWili to the uf2 bootloader.
 
@@ -707,7 +1096,7 @@ class FreeWili:
                 Processor to use.
 
         Returns:
-        -------
+        ---------
             Result[None, str]:
                 Returns Ok(None) if the command was sent successfully, Err(str) if not.
         """
@@ -738,12 +1127,12 @@ class FileMap:
         """Creates a FileMap from a file extension.
 
         Parameters:
-        ----------
+        ------------
             ext: str
                 File extension (ie. ".wasm"). Not case sensitive.
 
         Returns:
-        --------
+        ---------
             FileMap
 
         Raises:
@@ -756,8 +1145,8 @@ class FileMap:
             "wasm": (FreeWiliProcessorType.Main, "/scripts", "WASM binary"),
             "wsm": (FreeWiliProcessorType.Main, "/scripts", "WASM binary"),
             "zio": (FreeWiliProcessorType.Main, "/scripts", "ZoomIO script file"),
-            "bit": (FreeWiliProcessorType.Main, "/fpga", "FPGA bit file"),
-            "sub": (FreeWiliProcessorType.Display, "/radio", "Radio file"),
+            "bin": (FreeWiliProcessorType.Main, "/fpga", "FPGA bin file"),
+            "sub": (FreeWiliProcessorType.Main, "/radio", "Radio file"),
             "fwi": (FreeWiliProcessorType.Display, "/images", "Image file"),
             "wav": (FreeWiliProcessorType.Display, "/sounds", "Audio file"),
         }
@@ -770,12 +1159,12 @@ class FileMap:
         """Creates a FileMap from a file path.
 
         Parameters:
-        ----------
+        ------------
             file_name: str
                 File name (ie. "myfile.wasm"). Not case sensitive. Can contain paths.
 
         Returns:
-        --------
+        ---------
             FileMap
 
         Raises:
@@ -790,12 +1179,12 @@ class FileMap:
         """Creates a file path from the file_name to upload to the FreeWili.
 
         Parameters:
-        ----------
+        ------------
             file_name: str
                 File name (ie. "myfile.wasm"). Not case sensitive. Can contain paths.
 
         Returns:
-        --------
+        ---------
             str
                 Full file path intended to be uploaded to a FreeWili
 
